/***
 *
 * NNPDF C++ Driver
 *
 * Stefano Carrazza for the NNPDF Collaboration
 * email: stefano.carrazza@mi.infn.it
 *
 * August 2013
 *
 * Usage:
 *
 *  NNPDFDriver *pdf = new NNPDFDriver("gridname.LHgrid");
 *
 *  pdf->initPDF(0); // select replica [0,fMem]
 *
 *  pdf->xfx(x,Q,fl); // -> returns double
 *
 *  // with fl = [-6,7], LHAPDF format
 *
 *    or
 *
 *  pdf->xfx(x,Q); // -> returns double*
 *
 *  // with double* index from [0,13], always in the LHAPDF format
 *
 */

#include "NNPDFDriver.h"
#include <fstream>
#include <stdlib.h>
#include <cmath>

#define XMINGRID 1e-7
#define NNDriverVersion "1.0.3"

using namespace std;

/**
 * @brief Default constructor
 * @param gridfilename the string containing the LHgrid file name and location.
 */
NNPDFDriver::NNPDFDriver(string const& gridfilename):
  fNFL(13),
  fNX(100),
  fNQ2(50),
  fMem(1),
  fRep(0),
  fAlphas(0),
  fPDFGrid(NULL),
  fXGrid(NULL),
  fLogXGrid(NULL),
  fQ2Grid(NULL),
  fLogQ2Grid(NULL),
  fHasPhoton(false)
{
  // Logo
  cout << " ****************************************" << endl;
  cout << "      NNPDFDriver version " << NNDriverVersion << endl;
  cout << "  Grid: " << gridfilename << endl;
  cout << " ****************************************" << endl;

  // Read PDFs from file
  readPDFSet(gridfilename);
}

/**
 * @brief NNPDFDriver::~NNPDFDriver the destructor
 */
NNPDFDriver::~NNPDFDriver()
{
  if (fPDFGrid) {
    for (int imem = 0; imem <= fMem; imem++)
      {
        for (int i = 0; i < fNFL; i++)
          {
            for (int j = 0; j < fNX; j++)
              if (fPDFGrid[imem][i][j]) delete[] fPDFGrid[imem][i][j];
            if (fPDFGrid[imem][i]) delete[] fPDFGrid[imem][i];
          }
        if (fPDFGrid[imem]) delete[] fPDFGrid[imem];
      }
    delete[] fPDFGrid;
  }

  if (fXGrid) delete[] fXGrid;
  if (fLogXGrid) delete[] fLogXGrid;

  if (fQ2Grid) delete[] fQ2Grid;
  if (fLogQ2Grid) delete[] fLogQ2Grid;
}

/**
 * @brief NNPDFDriver::initPDF
 * @param irep
 */
void NNPDFDriver::initPDF(int irep)
{
  if (irep > fMem || irep < 0)
    {
      cout << "Error: replica out of range [0," << fMem << "]" << endl;
      exit(-1);
    }
  else
    fRep = irep;
}

/**
 * @brief NNPDFDriver::readPDFSet read the LHgrid file into an array
 * @param grid the LHgrid filename
 */
void NNPDFDriver::readPDFSet(string const& grid)
{
  fstream f;
  f.open(grid.c_str(),ios::in);
  if (f.fail())
    {
      cout << "Error: cannot open file " << grid << endl;
      exit(-1);
    }

  // Read header
  string tmp;
  for (;;) {
      getline(f,tmp);
      if (tmp.find("Alphas") != string::npos) break;
      cout << tmp << endl;
    }
  cout << endl;

  // Extract alphas and total number of members
  for (;;) {
      getline(f,tmp);
      if (tmp.find("Parameterlist:") != string::npos){
        f >> tmp >> fMem >> tmp >> tmp >> fAlphas;
        break;
        }
    }

  // Reading grid: removing header
  for (;;)
    {
      getline(f,tmp);
      if (tmp.find("NNPDF20intqed") != string::npos)
        {
          fHasPhoton = true;
          fNFL++;
          getline(f,tmp);
          break;
        }
      else if (tmp.find("NNPDF20int") != string::npos)
        {
          getline(f,tmp);
          break;
        }
    }

  // getting nx
  f >> fNX;
  fXGrid = new double[fNX];
  for (int ix = 0; ix < fNX; ix++)
    f >> fXGrid[ix];

  fLogXGrid = new double[fNX];
  for (int ix = 0; ix < fNX; ix++)
    fLogXGrid[ix] = log(fXGrid[ix]);

  f >> fNQ2;
  f >> tmp; // skip first value
  fQ2Grid = new double[fNQ2];
  for (int iq = 0; iq < fNQ2; iq++)
    f >> fQ2Grid[iq];

  fLogQ2Grid = new double[fNQ2];
  for (int iq = 0; iq < fNQ2; iq++)
    fLogQ2Grid[iq] = log(fQ2Grid[iq]);

  // Prepare grid array
  fPDFGrid = new double***[fMem+1];
  for (int imem = 0; imem <= fMem; imem++)
    {
      fPDFGrid[imem] = new double**[fNFL];
      for (int i = 0; i < fNFL; i++)
        {
          fPDFGrid[imem][i] = new double*[fNX];
          for (int j = 0; j < fNX; j++)
            {
              fPDFGrid[imem][i][j] = new double[fNQ2];
              for (int z = 0; z < fNQ2; z++)
                fPDFGrid[imem][i][j][z] = 0.0;
            }
        }
    }


  // read PDF grid points
  f >> tmp; // remove replica number
  for (int imem = 0; imem <= fMem; imem++)
    for (int ix = 0; ix < fNX; ix++)
      for (int iq = 0; iq < fNQ2; iq++)
        for (int fl = 0; fl < fNFL; fl++) 
          f >> fPDFGrid[imem][fl][ix][iq];	

  f.close();
}

/**
 * @brief NNPDFDriver::xfx
 * @param x
 * @param Q
 * @return
 */
double* NNPDFDriver::xfx(double const& x, double const& Q)
{
  return xfxevolve(x,Q);
}

/**
 * @brief NNPDFDriver::xfx
 * @param x
 * @param Q
 * @param id
 * @return
 */
double NNPDFDriver::xfx(double const& x, double const& Q, int const& id)
{
  return xfxevolve(x,Q,id+6)[0];
}

/**
 * @brief NNPDFDriver::xfxevolve
 * @param x
 * @param Q
 * @param id
 * @return
 */
double* NNPDFDriver::xfxevolve(double x, double Q, int id)
{
  double Q2 = Q*Q;

  // check bounds
  if (x < XMINGRID || x < fXGrid[0] || x > fXGrid[fNX-1]) {
    cout << "Parton interpolation: x out of range -- freezed" << endl;  
    if (x < fXGrid[0])  x = fXGrid[0];
    if (x < XMINGRID)   x = XMINGRID;
     if (x > fXGrid[fNX-1]) x = fXGrid[fNX-1];
  }
  if (Q2 < fQ2Grid[0] || Q2 > fQ2Grid[fNQ2-1]) {
    cout << "Parton interpolation: Q2 out of range -- freezed" << endl;  
    if (Q2 < fQ2Grid[0]) Q2 = fQ2Grid[0];
    if (Q2 > fQ2Grid[fNQ2-1]) Q2 = fQ2Grid[fNQ2-1];
  }

  // find nearest points in the grid
  int minx = 0;
  int maxx = fNX;

  while (maxx-minx > 1)
    {
      int midx = (minx+maxx)/2;
      if (x < fXGrid[midx]) 
	maxx = midx;
      else
	minx = midx;      
    }
  int ix = minx;

  int minq = 0;
  int maxq = fNQ2;
  while (maxq-minq > 1)
    {
      int midq = (minq+maxq)/2;
      if (Q2 < fQ2Grid[midq])
	maxq = midq;
      else
	minq = midq;      
    }
  int iq2 = minq;

  // Assign grid for interpolation. M,N -> order of polyN interpolation
  int   ix1a[fM], ix2a[fN];
  double x1a[fM], x2a[fN];
  double ya[fM][fN];

  for (int i = 0; i < fM; i++)
    {
      if (ix+1 >= fM/2 && ix+1 <= (fNX-fM/2)) ix1a[i] = ix+1 - fM/2 + i;
      if (ix+1 < fM/2) ix1a[i] = i;
      if (ix+1 > (fNX-fM/2)) ix1a[i] = (fNX-fM) + i;
	
      // Check grids
      if (ix1a[i] < 0 || ix1a[i] >= fNX)
	{
	  cout << "Error in grids! i, ixia[i] = " 
	       << i << "\t" << ix1a[i] << endl;
	  exit(-1);
	}
    }

  for (int j = 0; j < fN; j++)
    {
      if (iq2+1 >= fN/2 && iq2+1 <= (fNQ2-fN/2)) ix2a[j] = iq2+1 - fN/2 + j;
      if (iq2+1 < fN/2) ix2a[j] = j;
      if (iq2+1 > (fNQ2-fN/2)) ix2a[j] = (fNQ2-fN) + j;

      // Check grids
      if (ix2a[j] < 0 || ix2a[j] >= fNQ2)
	{
	  cout << "Error in grids! j, ix2a[j] = "
	       << j << "\t" << ix2a[j] << endl;
	  exit(-1);
	}
    }

  // define points where to evaluate interpolation
  // choose between linear or logarithmic (x,Q2) interpolation
  const double xch = 1e-1;

  double x1;
  if (x < xch) x1 = log(x);
  else x1 = x;
  double x2 = log(Q2);

  double *xpdf = new double[fNFL];
  for (int i = 0; i < fNFL; i++) xpdf[i] = 0.0;
  
  if (id < 0 || id > fNFL-1)
    {
      for (int ipdf = 0; ipdf < fNFL; ipdf++)
	{
	  // Choose betwen linear or logarithmic (x,Q2) interpolation
	  for (int i = 0; i < fM; i++)
	    {
	      if (x < xch)
		x1a[i] = fLogXGrid[ix1a[i]];
	      else
		x1a[i] = fXGrid[ix1a[i]];
	      
	      for (int j = 0; j < fN; j++)
		{
		  x2a[j] = fLogQ2Grid[ix2a[j]];
		  ya[i][j] = fPDFGrid[fRep][ipdf][ix1a[i]][ix2a[j]];
		}
	    }
	  
	  // 2D polynomial interpolation
	  double y = 0, dy = 0;
	  lh_polin2(x1a,x2a,ya,x1,x2,y,dy);
	  
	  xpdf[ipdf] = y;
	}
    }
  else
    {
      // Choose betwen linear or logarithmic (x,Q2) interpolation
      for (int i = 0; i < fM; i++)
	{
	  if (x < xch)
	    x1a[i] = fLogXGrid[ix1a[i]];
	  else
	    x1a[i] = fXGrid[ix1a[i]];
	  
	  for (int j = 0; j < fN; j++)
	    {
	      x2a[j] = fLogQ2Grid[ix2a[j]];
	      ya[i][j] = fPDFGrid[fRep][id][ix1a[i]][ix2a[j]];
	    }
	}
      
      // 2D polynomial interpolation
      double y = 0, dy = 0;
      lh_polin2(x1a,x2a,ya,x1,x2,y,dy);
      
      xpdf[0] = y;
    }

  return xpdf;
}

void NNPDFDriver::lh_polin2(double x1a[], double x2a[],
			    double ya[][fN],
			    double x1, double x2, 
			    double& y, double& dy)
{
  double yntmp[fN];
  double ymtmp[fM];

  for (int j = 0; j < fM; j++)
    {
      for(int k = 0; k < fN; k++)
	yntmp[k] = ya[j][k];
      
      lh_polint(x2a,yntmp,fN,x2,ymtmp[j],dy);
    }
  lh_polint(x1a,ymtmp,fM,x1,y,dy);
}

void NNPDFDriver::lh_polint(double xa[], double ya[], int n, double x,
			    double& y, double& dy)
{
  int ns = 0;  
  double dif = abs(x-xa[0]);
  double c[fM > fN ? fM : fN];
  double d[fM > fN ? fM : fN];
  
  for (int i = 0; i < n; i++)
    {
      double dift = abs(x-xa[i]);
      if (dift < dif)
	{
	  ns = i;
	  dif = dift;
	}
      c[i] = ya[i];
      d[i] = ya[i];
    }
  y = ya[ns];
  ns--;
  for (int m = 1; m < n; m++)
    {
      for (int i = 0; i < n-m; i++)
	{
	  double ho = xa[i]-x;
	  double hp = xa[i+m]-x;
	  double w = c[i+1]-d[i];
	  double den = ho-hp;
	  if (den == 0)
	    {
	      cout << "failure in polint" << endl;
	      exit(-1);	       
	    }
	  den = w/den;
	  d[i] = hp*den;
	  c[i] = ho*den;
	}
      if (2*(ns+1) < n-m)
	dy = c[ns+1];
      else {
	dy = d[ns];
	ns--;
      }
      y+=dy;
    }
}
